// Copyright 2017 Orus.io. All rights reserved.

package broker

import (
	"context"
	"fmt"

	nrpc "github.com/nats-rpc/gogo-nrpc"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/storage"
)

func (s *Service) receiveEnvelope(
	actorID api.UUID,
	context *api.ProcessingContext,
	output string,
	envelope *api.Envelope,
	close bool,
) (
	process *process.Process,
	sourceRef storage.NodeOutputRef,
	ack api.EnvelopeAck,
	err error,
) {

	actor, exists := s.Opts.ActorStorage.Get(actorID)
	if !exists {
		err = api.ClientProtoError("Unknown actor: %s", actorID)
		return
	}

	// If no context is given, make sure the sender is as emitter
	if (context == nil || context.ProcessID.IsZero()) &&
		actor.Kind != api.Actor_EMITTER {
		msg := fmt.Sprintf("Context-less envelope sent by non-emitter %s", actor.Name)
		s.Log.Errorf(msg)
		err = api.ClientProtoError(msg)
		return
	}

	// A context is given, make sure the process exists
	if context != nil && !context.ProcessID.IsZero() {
		process, err = s.processManager.Get(context.ProcessID)
		if err != nil {
			ack = *api.NewEnvelopeAckError(ack.ID, err)
			err = nil
			return
		}
		// TODO make sure NodeID matches the sending actor
		sourceRef.NodeID = context.NodeID
		sourceRef.Output = output
	}

	if close {
		ack.Status = api.EnvelopeAck_ACCEPTED

		process.UpdateReceiveStatus(
			api.UUID{}, actor.ID, sourceRef, storage.ReceptionClosed)
		return
	}

	if envelope != nil {
		s.Log.Debugf("Receiving envelope fragment (%s)", envelope.ID)
		ack.ID = envelope.ID
	} else {
		err = api.ClientProtoError("Missing envelope")
		s.Log.Errorf(err.Error())
		return
	}

	var (
		typesKnown bool
		eventTypes []string
	)

	// Check for empty envelope
	if envelope.Last == true && len(envelope.Events) == 0 && len(envelope.EventIDs) == 0 {
		if process == nil {
			ack = *api.NewEnvelopeAckError(
				envelope.ID,
				fmt.Errorf("Invalid envelop (missing or invalid context)"))
			return
		}

		// empty envelope, which can mean 2 things depending on the ID

		// Without an ID: the envelope is just a signal that the output
		// will not be used in this context. Inform the process.
		if envelope.ID.IsZero() {
			panic("Closing is not done here, should have done already")
		}
		// With an ID: the envelope is one sent on an input, hence must be
		// known to the broker which can reread it from storage.
		ack.Status = api.EnvelopeAck_ACCEPTED
		typesKnown = true
	} else {
		// non-empty envelope. Store it.
		var status storage.EnvelopeState

		for _, event := range envelope.Events {
			if t := event.GetType(); t != "" && !tools.ValidateEventType(t) {
				ack.Status = api.EnvelopeAck_ERROR
				ack.Reason = fmt.Sprintf("Invalid event type: '%s'", event.GetType())
				return
			}
		}

		status, err = s.Opts.EnvelopeStorage.StoreEnvelope(*envelope)
		if err != nil {
			panic(err)
		}
		s.Log.Debugf("Stored envelope: %s", status.Status)
		switch status.Status {
		case storage.EnvelopeStorageReceptionReceiving:
			ack.Status = api.EnvelopeAck_RECEIVING
		case storage.EnvelopeStorageReceptionComplete:
			ack.Status = api.EnvelopeAck_ACCEPTED
		case storage.EnvelopeStorageReceptionError:
			ack.Status = api.EnvelopeAck_ERROR
		default:
			ack.Status = api.EnvelopeAck_ERROR
		}
		typesKnown = status.TypesKnown
		eventTypes = status.EventTypes()

		if ack.Status == api.EnvelopeAck_ERROR {
			if process != nil {
				process.UpdateReceiveStatus(
					envelope.ID, actor.ID, sourceRef, storage.ReceptionError)
			}
			ack = *api.NewEnvelopeAckError(envelope.ID, status)
			return
		}
	}

	// The envelope is now safely stored
	// We just need the process to be aware of this reception, and we will
	// be done.

	if context == nil || context.ProcessID.IsZero() {
		process, err = s.processManager.LookupOrCreate(actor, envelope.ID)
		if err != nil {
			panic(fmt.Errorf("Error getting or creating process: %s", err))
		}
		if process.Pipeline != nil {
			sourceRef = process.SourceRef
		}
	}

	// No context but all types known -> the process can match a pipeline if needed
	if process.Pipeline == nil && (context == nil || context.ProcessID.IsZero()) && typesKnown {
		if matchErr := process.MatchPipeline(eventTypes); matchErr != nil {
			ack.Status = api.EnvelopeAck_ERROR
			ack.Reason = fmt.Sprintf(matchErr.Error())
			s.Log.Noticef(ack.Reason)
			return
		}
		sourceRef = process.SourceRef
	}

	// If process is unknown and the reception is over, we have a serious
	// problem
	if process == nil && ack.Status == api.EnvelopeAck_ACCEPTED {
		panic("Impossible state, please report")
	}

	if process != nil && process.Pipeline != nil {
		// If a process is associated to the envelope, make sure the route for the next
		// move is known, then update the reception status.
		// Logically we should calculate the route after updating the status, but if the
		// server crashes we would have route-less receptions that would be hard to
		// track and fix. Switching the two operations is harmless: a reception-less
		// route will become used as soon as the emitter re-attempt its sending.
		// And because the route is present, the emission service will find what
		// to do as soon as it is (re)started
		// See issue #39

		_, err := process.GetTargets(sourceRef, envelope.ID, false)
		if err != nil {
			panic(fmt.Sprintf("Error getting routes: %s", err))
		}

		var storageReceptionStatus storage.ReceptionStatus
		switch ack.Status {
		case api.EnvelopeAck_RECEIVING:
			storageReceptionStatus = storage.ReceptionOngoing
		case api.EnvelopeAck_ACCEPTED:
			storageReceptionStatus = storage.ReceptionDone
		}
		if err := process.UpdateReceiveStatus(
			envelope.ID, actor.ID, sourceRef, storageReceptionStatus,
		); err != nil {
			panic(fmt.Sprintf("Error updating receive status: %s", err))
		}
	}

	// our part (receive the envelope fragment) is done, we can safely
	// let the ack be sent to the sender

	return
}

type brokerServer struct {
	*Service
}

// Output implements the Actor.Output api
func (s brokerServer) Output(
	ctx context.Context,
	req api.OutputRequest,
) (
	api.EnvelopeAck, error,
) {
	s.TaskAdd()
	defer s.TaskDone()

	actorSID := nrpc.GetRequest(ctx).ServiceParam("actorID")
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return api.EnvelopeAck{}, err
	}

	_, _, ack, err := s.receiveEnvelope(
		actorID, req.Context, req.Output, req.Envelope, req.Close)

	return ack, err
}
