package core

import (
	"fmt"
	"time"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/storage"
)

type processLogEntries struct {
	processID  api.UUID
	logentries []storage.ProcessLogEntry
}

type processStatusChange struct {
	process storage.Process
	from    api.Process_Status
	to      api.Process_Status
}

func (s *Service) initProcessLogEntryHandler() error {
	// get the pending processing entries. If it fails we do nothing else
	logentries := s.o.ProcessStorage.GetUnprocessedProcessLogEntries()

	// Synchronously
	for _, entry := range logentries {
		changes, nodesReady := s.updateState(
			entry.ProcessID, []storage.ProcessLogEntry{entry})
		for _, nodesReady := range nodesReady {
			s.onNodesReady(nodesReady)
		}
		for _, change := range changes {
			s.onProcessStatusChange(change.process, change.from, change.to)
		}
	}

	// Init the routine
	s.TaskAdd()
	go func() {
		for entries := range s.processLogEntries {
			changes, nodesReady := s.updateState(entries.processID, entries.logentries)
			for _, nodesReady := range nodesReady {
				s.onNodesReady(nodesReady)
			}
			for _, change := range changes {
				s.onProcessStatusChange(change.process, change.from, change.to)
			}
		}
		s.TaskDone()
	}()

	return nil
}

func (s *Service) updateState(
	processID api.UUID, logentries []storage.ProcessLogEntry,
) (changes []processStatusChange, nodesReady []api.ProcessNodeReadyEvent) {
	p, ok := s.o.ProcessStorage.GetProcess(processID)
	if !ok {
		s.o.Logger.Errorf("No such process:  %s", processID)
		return
	}
	pipeline, ok := s.o.PipelineManager.GetPipeline(p.PipelineID)
	if !ok {
		s.o.Logger.Errorf("Process linked to unknown pipeline: %s", p.PipelineID)
		return
	}

	var unlocked bool
	stateUpdater := s.o.ProcessStorage.GetStateUpdater(processID)

	defer func() {
		if !unlocked {
			stateUpdater.Unlock()
		}
	}()

	// Calculate the next state
	pState, _ := stateUpdater.Get()

	var initialNodesReady = make(map[storage.ProcessNode]bool)
	for processNode, nodeState := range pState.NodeStates {
		if nodeState.Status == storage.ProcessNodeActive &&
			nodeState.Input == storage.ProcessNodeIOStarted { // TODO <=
			initialNodesReady[processNode] = true
		}
	}

	processError := func(entryIndex int, msg string, args ...interface{}) {
		msg = fmt.Sprintf(msg, args...)
		s.o.Logger.Errorf(msg)
		pState.Status = api.Process_ERROR
		pState.StatusReason = msg
		if err := stateUpdater.Set(pState); err != nil {
			s.o.Logger.Errorf("Could not save the process state: %s", err)
		}
		// Mark remaining logentries as error
		if entryIndex != -1 {
			for _, entry := range logentries[entryIndex:] {
				stateUpdater.MarkLogEntry(entry.ID, true)
			}
		}

		unlocked = true
		if err := stateUpdater.Unlock(); err != nil {
			s.o.Logger.Errorf("Could not save the process state: %s", err)
		}

		s.o.LogStorage.Log(api.LogEntry{
			ProcessID: processID,
			Message: api.LogMessage{
				Time:  time.Now(),
				Level: api.LogLevel_ERROR,
				Text:  msg,
			},
		})

	}

	statusList := []api.Process_Status{pState.Status}
	for i, logentry := range logentries {
		var err error
		pState, err = process.NextState(pipeline, pState, logentry, s.Log)
		if err != nil {
			processError(i, "Could not compute next state: %s", err)
			return
		}

		if pState.Status != statusList[len(statusList)-1] {
			statusList = append(statusList, pState.Status)
		}

		if err := stateUpdater.MarkLogEntry(logentry.ID, false); err != nil {
			processError(i, "Could not mark the entry: %s", err)
			return
		}
	}

	if err := stateUpdater.Set(pState); err != nil {
		processError(-1, "Could not save the state")
	}

	unlocked = true
	if err := stateUpdater.Unlock(); err != nil {
		s.o.Logger.Errorf("Could not unlock state: %s", err)
		return
	}

	// Emit a StatusChanged for each transition
	for i := range statusList[1:] {
		changes = append(changes, processStatusChange{p, statusList[i], statusList[i+1]})
	}

	// find all nodes that switched from notready to ready
	for processNode, nodeState := range pState.NodeStates {
		if nodeState.Status == storage.ProcessNodeActive &&
			nodeState.Input == storage.ProcessNodeIOStarted && // TODO <=
			!initialNodesReady[processNode] {
			nodesReady = append(nodesReady, api.ProcessNodeReadyEvent{
				ProcessID: processID,
				ActorID:   processNode.ActorID,
				NodeID:    processNode.NodeID,
			})
		}
	}

	return
}

func (s *Service) processLogEntryHandler(logentries []storage.ProcessLogEntry) {
	if len(logentries) != 0 {
		s.processLogEntries <- &processLogEntries{logentries[0].ProcessID, logentries}
	}
}
