package director

import (
	"errors"
	"sync"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

type input struct {
	job      *Job
	pushback bool
}

// JobQueue is a persistent queue for an actor jobs
type JobQueue struct {
	storage storage.JobStorage
	actorID api.UUID

	bufferSize int

	queuingFrozen  bool
	queuingFrozeAt storage.JobID

	in   chan input
	out  chan *Job
	done chan struct{}

	wg sync.WaitGroup
}

func newJobQueue(
	storage storage.JobStorage,
	actorID api.UUID,
	bufferSize int,
) *JobQueue {
	queue := JobQueue{
		storage:    storage,
		actorID:    actorID,
		bufferSize: bufferSize,
		in:         make(chan input),
		out:        make(chan *Job),
		done:       make(chan struct{}),
	}
	queue.wg.Add(1)
	ready := make(chan struct{})
	go queue.run(ready)
	<-ready
	return &queue
}

func (queue *JobQueue) run(ready chan<- struct{}) {
	defer queue.wg.Done()

	readyClosed := false
	defer func() {
		if !readyClosed {
			close(ready)
		}
	}()

	buffer := NewRingBuffer(queue.bufferSize)

	queueJob := func(in input) {
		if in.pushback {
			if buffer.Full() {
				buffer.TakeBack()
			}
			buffer.PushBack(in.job)
			if buffer.Full() && queue.queuingFrozen {
				queue.queuingFrozeAt = buffer.PeekLast().JobID
			}
		} else if !queue.queuingFrozen {
			buffer.Write(in.job)
			if buffer.Full() {
				queue.queuingFrozen = true
				queue.queuingFrozeAt = in.job.JobID
			}
		}
	}

	// Load the pending jobs
	{
		jobs := queue.storage.ListPendingJobByActor(
			queue.actorID, nil, buffer.Size)
		for _, job := range jobs {
			buffer.Write(&Job{Job: job})
		}
		if len(jobs) == buffer.Size {
			queue.queuingFrozen = true
			queue.queuingFrozeAt = jobs[len(jobs)-1].JobID
		}
	}

	readyClosed = true
	close(ready)

	var nextJob *Job
	var nextIsNil bool

	for {
		if !nextIsNil && nextJob == nil && !buffer.Empty() {
			nextJob = buffer.Read()
		}
		if !nextIsNil && nextJob == nil {
			select {
			case in := <-queue.in:
				if in.job == nil {
					nextIsNil = true
				} else {
					queueJob(in)
				}
			case <-queue.done:
				return
			}
		} else {
			select {
			case in := <-queue.in:
				if in.job == nil {
					nextIsNil = true
					if nextJob != nil {
						queueJob(input{nextJob, true})
						nextJob = nil
					}
				} else {
					queueJob(in)
				}
			case queue.out <- nextJob:
				if nextIsNil {
					nextIsNil = false
				}
				nextJob = nil
				// If half the buffer is free, load some data and try to unfreeze
				if queue.queuingFrozen && buffer.Count() < buffer.Size/2 {
					// Load data, and if possible, unfreeze
					max := buffer.Size - buffer.Count()
					jobs := queue.storage.ListPendingJobByActor(
						queue.actorID, &queue.queuingFrozeAt, max)
					for _, job := range jobs {
						buffer.Write(&Job{Job: job})
					}
					if len(jobs) < max {
						queue.queuingFrozen = false
					} else {
						queue.queuingFrozeAt = jobs[len(jobs)-1].JobID
					}
				}
			case <-queue.done:
				return
			}
		}
	}

}

// Close the queue resources and routine
func (queue *JobQueue) Close() {
	close(queue.done)
	queue.wg.Wait()
	close(queue.in)
	close(queue.out)
}

// Add a job to the queue. The job must already exist in the database
func (queue *JobQueue) Add(job *Job) {
	queue.in <- input{job, false}
}

// PushBack add a job in front of the queue
func (queue *JobQueue) PushBack(job *Job) {
	queue.in <- input{job, true}
}

// SendNilToOut sends a nil pointer to the output chan
func (queue *JobQueue) SendNilToOut() {
	queue.in <- input{}
}

// C returns a chan of jobs
func (queue *JobQueue) C() <-chan *Job {
	return queue.out
}

// ErrBufferFull is when the buffer is full
var ErrBufferFull = errors.New("RingBuffer is full")

// RingBuffer Structure
type RingBuffer struct {
	Size      int    // Size of the Ringbuffer
	Container []*Job // Array container of objects
	Reader    int    // Reader position
	Writer    int    // Writer Position
	empty     bool
	full      bool
}

// NewRingBuffer creates a new RingBuffer of initial size "size"
// Returns a pointer to the new RingBuffer
func NewRingBuffer(size int) *RingBuffer {
	rb := new(RingBuffer)
	rb.Size = size
	rb.Container = make([]*Job, size)
	rb.Reader = 0
	rb.Writer = 0
	rb.empty = true
	return rb
}

// Count returns the number of elements in the buffer
func (r *RingBuffer) Count() int {
	if r.empty {
		return 0
	}
	if r.full {
		return r.Size
	}
	w := r.Writer
	if w < r.Reader {
		w += r.Size
	}
	return w - r.Reader
}

// Empty returns true if the buffer is empty
func (r *RingBuffer) Empty() bool {
	return r.empty
}

// Full returns true if the buffer is full
func (r *RingBuffer) Full() bool {
	return r.full
}

// Write object into the RingBuffer
func (r *RingBuffer) Write(v *Job) error {
	if r.full {
		return ErrBufferFull
	}
	r.Container[r.Writer] = v
	r.Writer = (r.Writer + 1) % r.Size
	r.full = r.Writer == r.Reader
	r.empty = false
	return nil
}

// Read single object from the RingBuffer
func (r *RingBuffer) Read() (value *Job) {
	if r.empty {
		return nil
	}
	value = r.Container[r.Reader]
	r.Reader = (r.Reader + 1) % r.Size
	r.empty = r.Reader == r.Writer
	r.full = false
	return
}

// TakeBack removes the last inserted object
func (r *RingBuffer) TakeBack() (value *Job) {
	if r.empty {
		return nil
	}
	r.Writer = (r.Writer - 1) % r.Size
	value = r.Container[r.Writer]
	r.empty = r.Reader == r.Writer
	r.full = false
	return
}

// PushBack object in front of the buffer
func (r *RingBuffer) PushBack(v *Job) error {
	if r.full {
		return ErrBufferFull
	}

	r.Reader = (r.Reader - 1) % r.Size
	if r.Reader == -1 {
		r.Reader += r.Size
	}
	r.Container[r.Reader] = v
	r.empty = false
	r.full = r.Writer == r.Reader
	return nil
}

// PeekLast returns the last inserted object if any
func (r *RingBuffer) PeekLast() (value *Job) {
	if r.empty {
		return nil
	}
	index := (r.Writer - 1) % r.Size
	if index == -1 {
		index += r.Size
	}
	return r.Container[index]
}
