// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"os"

	"github.com/spf13/cobra"

	"bitbucket.org/orus-io/go-xbus"
	api "bitbucket.org/orus-io/go-xbus/api"
)

const hostnameFlag = "hostname"

// registerCmd represents the "register" command.
var registerCmd = &cobra.Command{
	Use:   "register",
	Short: "Generate a CSR and ask Xbus for a registration.",
	Long: `Generate a Certificate Signing Request then send it over to Xbus through the
unencrypted "registration" channel. This xbusctl will only be allowed to send
Xbus actual commands once accepted on the Xbus side.
This command also outputs the command one has to run on the Xbus side to accept
the registration request.`,
	Run: register,
}

// register parses arguments to proceed with the registration.
func register(cmd *cobra.Command, args []string) {
	os.Exit(doRegister(false, false))
}

// doRegister generates a CSR then asks Xbus for a registration.
func doRegister(renewKey, renewCert bool) int {
	xbusClient, err := GetClient(true)
	if err != nil {
		Logger.Errorf("%s", err)
		return 1
	}
	if err := xbusClient.Connect(); err != nil {
		Logger.Errorf("Cannot connect to server: %s", err)
		return 1
	}

	if err := xbusClient.CreateKey(renewKey); err != nil {
		if err != xbus.ErrAlreadyExists {
			Logger.Errorf("Cannot create a private key %s", err)
			return 1
		}
	}

	hostname := Viper.GetString("hostname")
	// TODO Read org and cn from the configuration
	if err := xbusClient.CreateCSR(hostname, "xbus", "xbusctl", renewCert); err != nil {
		if err != xbus.ErrAlreadyExists {
			Logger.Errorf("Cannot generate a CSR: %s", err)
			return 1
		}
	}

	// Send the registration request.
	regStatus, err := xbusClient.Register()
	if err != nil {
		Logger.Errorf("Cannot send the registration request: %s", err)
		return 1
	}

	Logger.Noticef("Successfully sent the registration request.")

	// Say something depending on the registration status we have received in reply.
	switch regStatus {
	case api.Registration_PENDING:
		msg := fmt.Sprintf(`The registration is pending approval. To approve from Xbus, run:

	xbusctl account accept %q
`,
			xbusClient.Name(),
		)
		for _, actor := range xbusClient.ActorOptions() {
			msg += fmt.Sprintf("\txbusctl actor accept %q\n", actor.Name)
		}

		if renewCert {
			msg += fmt.Sprintf(
				"\nThen run the 'register' command\n")
		} else {
			msg += fmt.Sprintf(
				"\nThen run the exact same command you just runned (register)\n")
		}
		Logger.Noticef(msg)
	case api.Registration_ACCEPTED:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically accepted.",
		)
	case api.Registration_REJECTED:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically rejected.",
		)
	}

	return 0
}

// renewCertCmd is the "renew-cert" command.
var renewCertCmd = &cobra.Command{
	Use:   "renew-cert",
	Short: "Re-generate a CSR and ask Xbus for its registration.",
	Long: `Re-generate a Certificate Signing Request and optionnaly a private key,
then do to register`,
	Run: func(cmd *cobra.Command, args []string) {
		renewKey := cmd.Flag("newkey").Value.String() == "true"
		os.Exit(doRegister(renewKey, true))
	},
}

func init() {
	RootCmd.AddCommand(registerCmd)
	flags := registerCmd.PersistentFlags()

	flags.StringP(
		hostnameFlag, "n", "localhost",
		"the hostname your client will use when connecting to the bus",
	)
	Viper.BindPFlag(hostnameFlag, flags.Lookup(hostnameFlag))

	RootCmd.AddCommand(renewCertCmd)
	flags = renewCertCmd.PersistentFlags()
	flags.Bool("newkey", false, "Also generate a new private key")
}
