// Copyright 2012-2015 Apcera Inc. All rights reserved.
// Copyright 2016 Orus.io for the modified parts
// this file is covered by the Apache 2.0 licence

// Package xbus provides a few things that all xbus actors (services & clients)
// may need
package xbus

import (
	"os"

	"bitbucket.org/orus-io/go-xbus/logger"
)

// LoggerOptions is the struct you need to pass to the GetLogger function
// it can be just created and passed untouched to the GetLogger if you need
// some quick logging functionality.
// it is recommended to at least set the Debug and Trace flags according to
// your taste though
type LoggerOptions struct {
	LogFile      string
	RemoteSyslog string
	Syslog       bool
	Logtime      bool
	Debug        bool
	Trace        bool
}

// Logger is the interface of xbus loggers
type Logger interface {
	// Log a notice statement
	Noticef(format string, v ...interface{})

	// Log a fatal error
	Fatalf(format string, v ...interface{})

	// Log an error
	Errorf(format string, v ...interface{})

	// Log a debug statement
	Debugf(format string, v ...interface{})

	// Log a trace statement
	Tracef(format string, v ...interface{})
}

// GetLogger is a convenience function that returns a logger configured
// according to your options and ready to be used in your server
func GetLogger(opts *LoggerOptions) Logger {
	var log Logger

	if opts.LogFile != "" {
		log = logger.NewFileLogger(opts.LogFile, opts.Logtime, opts.Debug, opts.Trace, true)
	} else if opts.RemoteSyslog != "" {
		log = logger.NewRemoteSysLogger(opts.RemoteSyslog, opts.Debug, opts.Trace)
	} else if opts.Syslog {
		log = logger.NewSysLogger(opts.Debug, opts.Trace)
	} else {
		colors := true
		// Check to see if stderr is being redirected and if so turn off color
		// Also turn off colors if we're running on Windows where os.Stderr.Stat() returns an invalid handle-error
		stat, err := os.Stderr.Stat()
		if err != nil || (stat.Mode()&os.ModeCharDevice) == 0 {
			colors = false
		}
		log = logger.NewStdLogger(opts.Logtime, opts.Debug, opts.Trace, colors, true)
	}
	return log
}

// NewPrefixedLogger PrefixedLogger constructor
func NewPrefixedLogger(prefix string, logger Logger) *PrefixedLogger {
	return &PrefixedLogger{
		prefix: prefix,
		logger: logger,
	}
}

// PrefixedLogger add a prefix to all the logs
type PrefixedLogger struct {
	prefix string
	logger Logger
}

// Noticef logs a notice statement
func (l *PrefixedLogger) Noticef(format string, v ...interface{}) {
	l.logger.Noticef(l.prefix+format, v...)
}

// Errorf logs an error statement
func (l *PrefixedLogger) Errorf(format string, v ...interface{}) {
	l.logger.Errorf(l.prefix+format, v...)
}

// Fatalf logs a fatal error
func (l *PrefixedLogger) Fatalf(format string, v ...interface{}) {
	l.logger.Fatalf(l.prefix+format, v...)
}

// Debugf logs a debug statement
func (l *PrefixedLogger) Debugf(format string, v ...interface{}) {
	l.logger.Debugf(l.prefix+format, v...)
}

// Tracef logs a trace statement
func (l *PrefixedLogger) Tracef(format string, v ...interface{}) {
	l.logger.Tracef(l.prefix+format, v...)
}

// NewNoopLogger returns a NoopLogger
func NewNoopLogger() NoopLogger { return NoopLogger{} }

// NoopLogger is a logger that does nothing
type NoopLogger struct{}

// Noticef logs a notice statement
func (NoopLogger) Noticef(format string, v ...interface{}) {}

// Errorf logs an error statement
func (NoopLogger) Errorf(format string, v ...interface{}) {}

// Fatalf logs a fatal error
func (NoopLogger) Fatalf(format string, v ...interface{}) {}

// Debugf logs a debug statement
func (NoopLogger) Debugf(format string, v ...interface{}) {}

// Tracef logs a trace statement
func (NoopLogger) Tracef(format string, v ...interface{}) {}
